% Functors:

    % cryptographic primitives
    fun sig/2.
    fun enc/2.

    % participant names 
    fun a/0.
    fun b/0.
    fun c/0. 
    fun d/0.
    fun intruder/0.

    % sessions
    fun s1/0.
    fun s2/0.

    % constants
    fun g/0.
    fun sec/0.
    fun ma/0.
    fun mb/0.
    fun md/0.
    fun first/0.
    fun second/0.

    % nonces and keys
    fun k/2.    % k(A,B) -- the long term key shared by A and B
    fun r/1.    % r(A) -- a random number generated by A 

    fun completedA/3.
    fun completedB/3.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    stages([0,1,2,3]).

%%%%% AUXILIARY PREDICATES %%%%%%%%%%%%

participant(a).
participant(b).
participant(c).
participant(d).
participant(intruder).
dishonest(intruder).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The secret:

    query sec.

% Intruders initial knowledge 

    rule :: g.
    rule :: A :- participant(A).
    rule :: k(A,B) :- participant(A), dishonest(B).
    rule :: k(B,A) :- participant(A), dishonest(B).

% Intruder rules

    rule :: K, M -> enc(K,M).
    rule :: K, enc(K,M) -> M.

% Protocol rules

    % A:
    rule::     g,  g^r(a).

    % B:
    rule(1)::  Xa, Xb -> Xa^r(b), Xb, Xb^r(b).

    % C:
    rule(2)::  Xa, Xb, Xc -> Xa^r(c), Xb^r(c), Xc, Xc^r(c).
            
    % D:
    rule(3)::  Xa, Xb, Xc, Xd  ->   Xa^r(d)^k(a,d),
                                    Xb^r(d)^k(b,d),
                                    Xc^r(d)^k(c,d),
                                    enc(Xd^r(d), sec).

    % completed A:
    rule :: X -> enc(X^r(a)^~k(a,d), sec).

    % completed B:
    rule :: X -> enc(X^r(b)^~k(b,d), sec).
    
    % completed C:
    rule :: X -> enc(X^r(c)^~k(c,d), sec).


%%%%%% TESTS %%%%%%%
    
    % SHOULD FIRE:
    
    % sent to A:
    % rule :: g^r(b)^r(c)^r(d)^k(a,d) -> sec.    

    % sent to B:
    % rule :: g^r(a)^r(c)^r(d)^k(c,d) -> sec.    

    % sent to c:
    % rule :: g^r(a)^r(b)^r(d)^k(c,d) -> sec.    
    

    % rule :: enc(g^r(a)^r(b)^r(c)^r(d), sec) -> sec.

