% Functors:

    % cryptographic primitives
    fun sig/2.
    fun enc/2.

    % participant names 
    fun a/0.
    fun b/0.
    fun c/0. % the intruder
    fun d/0.

    % sessions
    fun s1/0.
    fun s2/0.

    % constants
    fun g/0.
    fun sec/0.
    fun sec_a/0.
    fun sec_b/0.
    fun sec_d/0.
    fun ma/0.
    fun mb/0.
    fun md/0.
    fun first/0.
    fun second/0.

    % nonces and keys
    fun k/2.    % k(A,B) -- the long term key shared by A and B
    fun r/2.    % r(A,S) -- a random number generated by A in session S

    fun completedA/3.
    fun completedB/3.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    stages([0,1,2]).

%%%%% AUXILIARY PREDICATES %%%%%%%%%%%%

participant(a).
participant(b).
participant(c).
participant(d).
honest(a).
honest(b).
honest(d).
dishonest(c).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The secret:

    query sec.

% Intruders initial knowledge 

    rule :: g.
    rule :: A :- participant(A).
    rule :: k(A,B) :- participant(A), dishonest(B).
    rule :: k(B,A) :- participant(A), dishonest(B).

% Intruder rules

    rule :: K, M -> enc(K,M).
    rule :: K, enc(K,M) -> M.

% Protocol rules

%%%% SESSION 1

    % A:
    rule::   g,  g^r(a,s1).

    % B:
    rule(1):: Xa, Xb -> Xa^r(b,s1), Xb, Xb^r(b,s1).

            
    % D:
    rule(2)::  Xa, Xb, Xc, Xd  ->   Xa^r(d,s1)^k(a,d),
                                    Xb^r(d,s1)^k(b,d),
                                    Xc^r(d,s1)^k(c,d),
                                    enc(Xd^r(d,s1), md).

    % completed a:
    % rule :: X -> enc(X^r(a,s1)^~k(a,d), ma).

    % completed b:
    % rule :: X -> enc(X^r(a,s1)^~k(a,d), mb).

%%%% SESSION 2

    % A:
    rule::  g, g^r(a,s2).

    % B:
    rule(1):: Xa, Xb -> Xa^r(b,s2), Xb, Xb^r(b,s2).

    % D:
    rule(2)::  Xa, Xb, Xd  ->   Xa^r(d,s2)^k(a,d),
                                    Xb^r(d,s2)^k(b,d),
                                    enc(Xd^r(d,s2), md).

    % completed a:
    rule :: X -> enc(X^r(a,s2)^~k(a,d), sec).

    % completed b:
    % rule :: X -> enc(X^r(b,s3)^~k(b,d), sec).

%%%%%% TESTS %%%%%%%
    
    % SHOULD FIRE:
    
    % The intruder can read messages encrypted under the key derived
    % from the first session:
    % rule :: ma,mb,md -> sec.
    
    % The intruder knows the key derived from the first session:
    % rule :: g^r(a,s1)^r(b,s1)^r(d,s1) -> sec.    


    % sent to a:
    %rule :: g^r(b,s1)^r(d,s1)^k(a,d) -> sec.    

    % sent to b:
    % rule :: g^r(a,s1)^r(d,s1)^k(c,d) -> sec.    

    % sent to c:
    % rule :: g^r(a,s1)^r(b,s1)^r(d,s1)^k(c,d) -> sec.    
    
    % rule(2) :: g^r(a,s2) -> sec.

%%%%

    % A has completed the second session:
    % rule :: enc(g^r(a,s2)^r(b,s2)^r(d,s2), sec_a) -> sec.

    % B has completed the second session:
    % rule :: enc(g^r(a,s2)^r(b,s2)^r(d,s2), sec_b) -> sec.
    
