%%
%%  IKEv2-Child
%%  
%%  Modelling follows the one from the AVISPA Library of Protocols.
%%


% Functors:

    % cryptographic primitives
    fun sig/2.
    fun enc/2.
    fun f/1.

    % participant names 
    fun a/0.
    fun b/0.
    fun c/0.

    % constants
    fun g/0.
    fun sec/0.
    fun zero/0.
    fun one/0.

    % nonces and keys
    fun k/1.    % k(K) -- private key of X.
    fun pub/1.

    fun psk/2.  % psk(X,Y) -- a symmetric shared by X and Y.

    fun ni/2.  % ...(X,Y) -- a nonce generated by X in the session with Y.
    fun nr/2.  
    fun dhx/2. 
    fun dhy/2.
    fun sa/2.
    fun ma/2.
    fun mb/2.

    fun completedA/6.
    fun completedB/6.


%%%%% AUXILIARY PREDICATES %%%%%%%%%%%%

participant(a).
participant(b).
participant(c).

%%
% The following predicates specify which session will be
% taken into consideration.

% roleA(X,Y)  ---  X as an initiator runs the protocol with Y as responder 
% roleB(X,Y)  ---  X as a responder runs the protocol with Y as initiator 

roleA(a, a, psk(a,a)).
% roleA(a, b, psk(a,b)).
% roleA(b, a, psk(a,b)).
roleA(a, c, psk(a,c)).
% roleA(b, c, psk(b,c)).

roleB(a, a, psk(a,a)).
% roleB(a, b, psk(a,b)).
% roleB(b, a, psk(a,b)).
roleB(a, c, psk(a,c)).
% roleB(b, c, psk(b,c)).

% roleA( A, B, SK, X, NI, SA, MA )
%
roleA( A, B, SK, dhx(A,B), ni(A,B), sa(A,B), ma(A,B) ) :- roleA(A,B,SK).
    
% roleB( B, A, SK, Y, NR, MB )
%
roleB(B, A, SK, dhy(B,A), nr(B,A), mb(B,A)) :- roleB(B,A,SK).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The secret:

    query sec.

% Intruder rules

    rule:: sig(_,M) -> M.
    rule:: sig(K,_) -> pub(K).
    rule:: K, M -> sig(K,M).
    rule:: K, M -> enc(K,M).
    rule:: K, enc(K,M) -> M.
    rule:: M -> f(M).

    % rule:: X -> km(X).
    % rule:: X -> ke(X).
    % rule:: X -> ks(X).

% Initial intruder knowledge

    rule:: X             :- participant(X).
    rule:: pub(k(X))     :- participant(X).
    rule:: k(c).
    rule:: g.

    rule:: psk(c,X)        :- participant(X).
    rule:: psk(X,c)        :- participant(X).

% Protocol Rules (the main session):


    % (1) A->B
    %
    rule::
        enc(SK, (SA, NI, g^X))
    :-
        roleA( _A, _B, SK, X, NI, SA, _MA ).

    % (2) B->A
    %
    rule::
        enc(SK, (SA, _NI, _GX))
      ->
        enc(SK, (SA, NR, g^Y))
    :-
        roleB( _B, _A, SK, Y, NR, _MB ).


    % (3) A->B
    %
    rule::
        enc(SK, (SA, NR, GY))
      ->
        enc(CSK, (MA, zero))
    :-
        roleA( _A, _B, SK, X, NI, SA, MA ),
        CSK = f((NI, NR, SA, GY^X)).


    % (4) B->A
    %
    rule::
        enc(SK, (SA, NI, GX)), % from the previous step
        enc(CSK, (_MA, zero))
      ->
        enc(CSK, (MB, one)),
        completedB(B,A,GX,g^Y,GX^Y,CSK)
    :-
        roleB( B, A, SK, Y, NR, MB ),
        CSK = f((NI, NR, SA, GX^Y)).

    % (5) A completed
    %
    rule::
        enc(SK, (SA, NR, GY)), % from the previous step
        enc(CSK, (_MB, one))
      ->
        completedA(A, B, g^X, GY, GY^X, CSK)
    :-
        roleA( A, B, SK, X, NI, SA, _MA ),
        CSK = f((NI, NR, SA, GY^X)).


%%%%%%%%%%%%%%%%%%%%

    rule:: SK, completedA(a,a,_,_,_,SK) -> sec.
    rule:: SK, completedB(a,a,_,_,_,SK) -> sec.

    rule:: GXY, completedA(a,a,_,_,GXY,_) -> sec.
    rule:: GXY, completedB(a,a,_,_,GXY,_) -> sec.

    rule:: g^dhx(a,a)^dhy(a,a) -> sec.
    rule:: g^dhx(a,a)^dhy(a,a) -> sec.

    rule:: completedA(a, a, _, g^dhy(a,c), _, _) -> sec.
    rule:: completedB(a, a, g^dhx(a,c), _, _, _) -> sec. 

   
    %%%
    %% should fire
    %
    % rule:: completedA(a,a,GX,GY,GXY,SK), completedB(a,a,GX,GY,GXY,SK) -> sec.
    % rule:: SK, completedA(a,c,_,_,_,SK) -> sec.
    % rule:: SK, completedB(a,c,_,_,_,SK) -> sec.
    
